#version 110


#extension GL_ARB_shading_language_include : require


#include <tests/all_topdir.h>
#include "all_topdir.h"
#include <tests/all_subdir/all_subdir.h>
#include "all_subdir/all_subdir.h"
#include </tests/all_topdir.h>
#include "/all_topdir.h"
#include </tests/all_subdir/all_subdir.h>
#include "/all_subdir/all_subdir.h"
#include "all_recursive.h"
#include <tests/all_recursive.h>
#include "all_subdir/all_recursive.h"
#include <tests/all_subdir/all_recursive.h>


attribute vec4 vert_position;


void main()
{
    gl_Position = gl_Vertex + vert_position;
}
