#ifndef SUBDIR_RECURSIVE_H_
#define SUBDIR_RECURSIVE_H_


#include <tests/all_subdir/all_recursive.h>
#include "../all_subdir/all_recursive.h"
#include <../../../../../tests/all_subdir/all_recursive.h>
#include "../../../../../tests/all_subdir/all_recursive.h"


#endif // SUBDIR_RECURSIVE_H_
